import einops
import torch
from torch import nn


class VitSeperateNorm(nn.Module):
    def __init__(
            self,
            dim,
            num_aux_tokens=1,
            aux_norm=nn.LayerNorm,
            patch_norm=nn.LayerNorm,
            affine=True,
            eps=1e-6,
    ):
        assert num_aux_tokens > 0
        super().__init__()
        self.num_aux_tokens = num_aux_tokens
        self.aux_norm = self._instantiate_norm(dim=dim * num_aux_tokens, ctor=aux_norm, eps=eps, affine=affine)
        self.patch_norm = self._instantiate_norm(dim=dim, ctor=patch_norm, eps=eps, affine=affine)

    @staticmethod
    def _instantiate_norm(ctor, dim, affine, eps):
        if ctor == nn.LayerNorm:
            return nn.LayerNorm(dim, eps=eps, elementwise_affine=affine)
        if ctor == nn.BatchNorm1d:
            return nn.BatchNorm1d(dim, eps=eps, affine=affine)
        raise NotImplementedError

    def forward(self, x):
        if self.num_aux_tokens == 0:
            return self.patch_norm(x)

        # norm aux tokens (multiple token case: concat aux tokens -> batchnorm -> split)
        flat_aux_tokens = einops.rearrange(
            x[:, :self.num_aux_tokens],
            "batch_size num_aux_tokens dim -> batch_size (num_aux_tokens dim)",
        )
        aux_normed = einops.rearrange(
            self.aux_norm(flat_aux_tokens),
            "batch_size (num_aux_tokens dim) -> batch_size num_aux_tokens dim",
            num_aux_tokens=self.num_aux_tokens,
        )

        patch_tokens = x[:, self.num_aux_tokens:]
        patch_normed = self.patch_norm(patch_tokens)
        return torch.concat([aux_normed, patch_normed], dim=1)



from .vit_batchnorm import VitBatchNorm1d, VitBatchNorm2d, VitBatchNorm3d
from .vit_block import VitBlock
from .vit_mlp import VitMlp
from .vit_patch_embed import VitPatchEmbed
from .vit_seperate_norm import VitSeperateNorm
from .vit_pos_embed import VitPosEmbed1d, VitPosEmbed2d, VitPosEmbed3d
from .vit_class_tokens import VitClassTokens


import torch
from torch import nn

from kappamodules.functional.pos_embed import (
    get_sincos_1d_from_seqlen,
    get_sincos_2d_from_seqlens,
    get_sincos_3d_from_seqlens,
)


class VitPosEmbed1d(nn.Module):
    def __init__(self, seqlen: int, dim: int, is_learnable=False):
        super().__init__()
        self.seqlen = seqlen
        self.dim = dim
        self.is_learnable = is_learnable
        if is_learnable:
            self.embed = nn.Parameter(torch.zeros(1, seqlen, dim))
        else:
            self.register_buffer("embed", get_sincos_1d_from_seqlen(seqlen=seqlen, dim=dim).unsqueeze(0))
        self.reset_parameters()

    def reset_parameters(self):
        if self.is_learnable:
            nn.init.trunc_normal_(self.embed, std=.02)

    def forward(self, x):
        assert x.ndim == 3
        return x + self.embed


class VitPosEmbed2d(nn.Module):
    def __init__(self, seqlens: int, dim: int, is_learnable=False):
        super().__init__()
        self.seqlens = seqlens
        self.dim = dim
        self.is_learnable = is_learnable
        if is_learnable:
            self.embed = nn.Parameter(torch.zeros(1, *seqlens, dim))
        else:
            self.register_buffer("embed", get_sincos_2d_from_seqlens(seqlens=seqlens, dim=dim).unsqueeze(0))
        self.reset_parameters()

    def reset_parameters(self):
        if self.is_learnable:
            nn.init.trunc_normal_(self.embed, std=.02)

    def forward(self, x):
        assert x.ndim == 4
        return x + self.embed


class VitPosEmbed3d(nn.Module):
    def __init__(self, seqlens: int, dim: int, is_learnable=False):
        super().__init__()
        self.seqlens = seqlens
        self.dim = dim
        self.is_learnable = is_learnable
        if is_learnable:
            self.embed = nn.Parameter(torch.zeros(1, *seqlens, dim))
        else:
            self.register_buffer("embed", get_sincos_3d_from_seqlens(seqlens=seqlens, dim=dim).unsqueeze(0))
        self.reset_parameters()

    def reset_parameters(self):
        if self.is_learnable:
            nn.init.trunc_normal_(self.embed, std=.02)

    def forward(self, x):
        assert x.ndim == 5
        return x + self.embed



import torch



def get_sincos_1d_from_seqlen(seqlen: int, dim: int):
    grid = torch.arange(seqlen, dtype=torch.double)
    return get_sincos_1d_from_grid(grid=grid, dim=dim)

def get_sincos_1d_from_grid(grid, dim: int):
    if dim % 2 == 0:
        padding = None
    else:
        padding = torch.zeros(*grid.shape, 1)
        dim -= 1
    # generate frequencies for sin/cos (e.g. dim=8 -> omega = [1.0, 0.1, 0.01, 0.001])
    omega = 1. / 10000 ** (torch.arange(0, dim, 2, dtype=torch.double) / dim)
    # create grid of frequencies with timesteps
    # Example seqlen=5 dim=8
    # [0, 0.0, 0.00, 0.000]
    # [1, 0.1, 0.01, 0.001]
    # [2, 0.2, 0.02, 0.002]
    # [3, 0.3, 0.03, 0.003]
    # [4, 0.4, 0.04, 0.004]
    # Note: supports cases where grid is more than 1d
    out = grid.unsqueeze(-1) @ omega.unsqueeze(0)
    emb_sin = torch.sin(out)
    emb_cos = torch.cos(out)
    emb = torch.concat([emb_sin, emb_cos], dim=-1).float()
    if padding is None:
        return emb
    else:
        return torch.concat([emb, padding], dim=-1)

def get_sincos_2d_from_seqlens(seqlens, dim: int):
    seqlen_h, seqlen_w = seqlens
    grid_h = torch.arange(seqlen_h, dtype=torch.double)
    grid_w = torch.arange(seqlen_w, dtype=torch.double)
    grid = torch.meshgrid(grid_h, grid_w, indexing="xy")
    grid = torch.stack(grid).reshape(2, seqlen_h, seqlen_w)
    return get_2d_sincos_pos_embed_from_grid(grid=grid, dim=dim)

def get_2d_sincos_pos_embed_from_grid(grid, dim: int):
    assert dim % 2 == 0
    grid_h, grid_w = grid
    emb_h = get_sincos_1d_from_grid(grid=grid_h, dim=dim // 2)
    emb_w = get_sincos_1d_from_grid(grid=grid_w, dim=dim // 2)
    return torch.concat([emb_h, emb_w], dim=-1)

def get_sincos_3d_from_seqlens(seqlens, dim: int):
    seqlen_x, seqlen_y, seqlen_z = seqlens
    grid_x = torch.arange(seqlen_x, dtype=torch.double)
    grid_y = torch.arange(seqlen_y, dtype=torch.double)
    grid_z = torch.arange(seqlen_z, dtype=torch.double)
    grid = torch.meshgrid(grid_x, grid_y, grid_z, indexing="xy")
    grid = torch.stack(grid).reshape(3, seqlen_x, seqlen_y, seqlen_z)
    return get_3d_sincos_pos_embed_from_grid(grid=grid, dim=dim)

def get_3d_sincos_pos_embed_from_grid(grid, dim: int):
    assert dim % 3 == 0
    grid_x, grid_y, grid_z = grid
    emb_x = get_sincos_1d_from_grid(grid=grid_x, dim=dim // 3)
    emb_y = get_sincos_1d_from_grid(grid=grid_y, dim=dim // 3)
    emb_z = get_sincos_1d_from_grid(grid=grid_z, dim=dim // 3)
    return torch.concat([emb_x, emb_y, emb_z], dim=-1)

def get_sincos_pos_embed_from_seqlens(seqlens, dim: int):
    assert isinstance(seqlens, (tuple, list))
    ndim = len(seqlens)
    grids = [torch.arange(seqlen, dtype=torch.double) for seqlen in seqlens]
    grid = torch.meshgrid(*grids, indexing="xy")
    grid = torch.stack(grid).reshape(ndim, *seqlens)
    return get_sincos_pos_embed_from_grid(grid=grid, dim=dim)

def get_sincos_pos_embed_from_grid(grid, dim: int):
    ndim = grid.size(0)
    if dim % ndim == 0:
        padding = None
    else:
        padding_dim = dim % ndim
        padding = torch.zeros(*grid.shape[1:], padding_dim)
        dim -= padding_dim
    pos_embed = torch.concat([get_sincos_1d_from_grid(grid=grid[i], dim=dim // ndim) for i in range(ndim)], dim=-1)
    if padding is None:
        return pos_embed
    else:
        return torch.concat([pos_embed, padding], dim=-1)




import einops
from torch import nn


class VitBatchNorm1d(nn.BatchNorm1d):
    def forward(self, x):
        if x.ndim == 3:
            seqlen = x.size(1)
            x = einops.rearrange(x, "b l c -> (b l) c")
            x = super().forward(x)
            x = einops.rearrange(x, "(b l) c -> b l c", l=seqlen)
        elif x.ndim == 2:
            # support single token as well (e.g. if norm is applied after pooling)
            x = super().forward(x)
        else:
            raise RuntimeError(
                f"expected 3d (batch_size, seqlen, dim) or "
                f"2d (batch_size, dim) input got {tuple(x.shape)}"
            )
        return x


class VitBatchNorm2d(nn.BatchNorm1d):
    def forward(self, x):
        if x.ndim == 4:
            _, _, height, width = x.shape
            x = einops.rearrange(x, "b h w c -> (b h w) c")
            x = super().forward(x)
            x = einops.rearrange(x, "(b h w) c -> b h w c", l=seqlen)
        elif x.ndim == 2:
            # support single token as well (e.g. if norm is applied after pooling)
            x = super().forward(x)
        else:
            raise RuntimeError(
                f"expected 4d (batch_size, height, width, dim) or "
                f"2d (batch_size, dim) input got {tuple(x.shape)}"
            )
        return x


class VitBatchNorm3d(nn.BatchNorm1d):
    def forward(self, x):
        if x.ndim == 5:
            _, _, height, width = x.shape
            x = einops.rearrange(x, "b x y z c -> (b x y z) c")
            x = super().forward(x)
            x = einops.rearrange(x, "(b x y z) c -> b x y z c", l=seqlen)
        elif x.ndim == 2:
            # support single token as well (e.g. if norm is applied after pooling)
            x = super().forward(x)
        else:
            raise RuntimeError(
                f"expected 5d (batch_size, x, y, z, dim) or "
                f"2d (batch_size, dim) input got {tuple(x.shape)}"
            )
        return x



import einops
import torch.nn.functional as F
from torch import nn

from kappamodules.init import initialize_xavier_uniform_zero_bias, initialize_qkv_seperately


class DotProductAttention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=True):
        super().__init__()
        assert hasattr(F, "scaled_dot_product_attention")
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.num_heads = num_heads
        self.head_dim = dim // num_heads

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)

    def reset_parameters(self):
        self.apply(initialize_xavier_uniform_zero_bias)
        initialize_qkv_seperately(self)

    def _forward(self, x):
        q, k, v = einops.rearrange(
            self.qkv(x),
            "bs seqlen (three num_heads head_dim) -> three bs num_heads seqlen head_dim",
            three=3,
            num_heads=self.num_heads,
            head_dim=self.head_dim,
        ).unbind(0)
        x = F.scaled_dot_product_attention(q, k, v)
        x = einops.rearrange(x, "bs num_heads seqlen head_dim -> bs seqlen (num_heads head_dim)")
        x = self.proj(x)
        return x

    def forward(self, x):
        raise NotImplementedError


class DotProductAttention1d(DotProductAttention):
    def forward(self, x):
        return self._forward(x)


class DotProductAttention2d(DotProductAttention):
    def forward(self, x):
        _, _, h, w = x.shape
        x = einops.rearrange(x, "b c h w -> b (h w) c")
        x = self._forward(x)
        x = einops.rearrange(x, "bs (h w) dim -> bs dim h w", h=h, w=w)
        return x


class DotProductAttention3d(DotProductAttention):
    def forward(self, x):
        _, _, h, w, d = x.shape
        x = einops.rearrange(x, "b c h w d -> b (h w d) c")
        x = self._forward(x)
        x = einops.rearrange(x, "bs (h w d) dim -> bs dim h w d", h=h, w=w, d=d)
        return x
